/* Copyright (c) 2011-2017, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the OSU HiBD-Benchmarks software package
 * developed by the team members of The Ohio State University's
 * Network-Based Computing Laboratory (NBCL), headed by Professor
 * Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to
 * the copyright file COPYRIGHT in the top level directory.
 *
 */
/**
 * Copyright (c) 2014-2015 YCSB contributors. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * permissions and limitations under the License. See accompanying
 * LICENSE file.
 */

package com.yahoo.ycsb.db;

import com.yahoo.ycsb.ByteIterator;
import com.yahoo.ycsb.DB;
import com.yahoo.ycsb.DBException;
import com.yahoo.ycsb.Status;
import com.yahoo.ycsb.StringByteIterator;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.net.InetSocketAddress;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

// Default Memcached client using spymemcached
import net.spy.memcached.ConnectionFactoryBuilder;
import net.spy.memcached.FailureMode;
// We also use `net.spy.memcached.MemcachedClient`; it is not imported
// explicitly and referred to with its full path to avoid conflicts with the
// class of the same name in this file.
import net.spy.memcached.internal.OperationFuture;
import net.spy.memcached.internal.*;

import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonGenerator;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.node.ObjectNode;
import org.apache.log4j.Logger;

// RDMA-Memcached Client using RDMA-Libmecached library
// wrapped via JNI through RdmaMemcachedClient class
import edu.osu.hibd.ohb.memcached.RdmaMemcachedClient;

import static java.util.concurrent.TimeUnit.MILLISECONDS;

/**
 * Memcached client implementation.
 */
public class MemcachedClient extends DB {

  private final Logger logger = Logger.getLogger(getClass());
  protected static final ObjectMapper MAPPER = new ObjectMapper();

  private boolean isCopyJni = false; 
  private boolean checkOperationStatus;
  private long shutdownTimeoutMillis;
  private int objectExpirationTime;
  private int optMode; 

  /** The name of the property for the server cluster. */
  public static final String HOSTS_PROPERTY = "memcached.hosts";
  /** The name of the property for the key's maximum size in bytes. */
  public static final String KEY_MAXSIZE_PROPERTY = "memcached.key.maxsize";
  public static final String DEFAULT_KEY_MAXSIZE = "256"; 
  /** The name of the property for number of native buffers for sharing data. */
  public static final String NATIVEBUFFER_NUM_PROPERTY = "memcached.native.count";
  public static final String DEFAULT_NATIVEBUFFER_NUM = "2"; 
  /** Default port number. */
  public static final int DEFAULT_PORT = 11211;
  public static final int JSON_FORMAT_EXTRA = 256; 

  /** The name of the property for the number of fields in a record. */
  public static final String FIELD_COUNT_PROPERTY = "fieldcount";

  /** Default number of fields in a record. */
  public static final String FIELD_COUNT_PROPERTY_DEFAULT = "10";

  /** The name of the property for the field length in a record. */
  public static final String FIELD_LENGTH_PROPERTY = "fieldlength";

  /** Default fields length in a record. */
  public static final String FIELD_LENGTH_PROPERTY_DEFAULT = "100";

  private static final String TEMPORARY_FAILURE_MSG = "Temporary failure";
  private static final String CANCELLED_MSG = "cancelled";

  public static final String SHUTDOWN_TIMEOUT_MILLIS_PROPERTY =
        "memcached.shutdownTimeoutMillis";
  public static final String DEFAULT_SHUTDOWN_TIMEOUT_MILLIS = "30000";

  public static final String OBJECT_EXPIRATION_TIME_PROPERTY =
        "memcached.objectExpirationTime";
  public static final String DEFAULT_OBJECT_EXPIRATION_TIME =
        String.valueOf(Integer.MAX_VALUE);

  public static final String CHECK_OPERATION_STATUS_PROPERTY =
        "memcached.checkOperationStatus";
  public static final String CHECK_OPERATION_STATUS_DEFAULT = "true";

  public static final String READ_BUFFER_SIZE_PROPERTY =
        "memcached.readBufferSize";
  public static final String DEFAULT_READ_BUFFER_SIZE = "3000000";

  public static final String OP_TIMEOUT_PROPERTY = "memcached.opTimeoutMillis";
  public static final String DEFAULT_OP_TIMEOUT = "60000";

  public static final String FAILURE_MODE_PROPERTY = "memcached.failureMode";
  public static final FailureMode FAILURE_MODE_PROPERTY_DEFAULT =
        FailureMode.Redistribute;

  public static final String PROTOCOL_PROPERTY = "memcached.protocol";
  public static final String DEFAULT_PROTOCOL = "RDMA"; 
  public static final ConnectionFactoryBuilder.Protocol 
        DEFAULT_SPYMEMCACHED_PROTOCOL = ConnectionFactoryBuilder.Protocol.TEXT; 

  /**
   * The MemcachedClient implementation that will be used to communicate
   * with the memcached server.
   */
  private net.spy.memcached.MemcachedClient defClient;
  private RdmaMemcachedClient client;

  /**
   * @returns Underlying Memcached protocol client, implemented by
   *     SpyMemcached.
   */
  protected net.spy.memcached.MemcachedClient memcachedClient() {
    return defClient;
  }

  @Override
  public void init() throws DBException {
    String protocolString = getProperties().getProperty(PROTOCOL_PROPERTY); 

    try {
      if (protocolString.equals("RDMA")) {
        // ***** using RDMA-Libmemcached + Blocking mode as default
        String hosts = getProperties().getProperty(HOSTS_PROPERTY);

        // ***** Create a list of requests; size is max. outstanding requests
        int maxKeyLength = Integer.parseInt(
            getProperties().getProperty(KEY_MAXSIZE_PROPERTY,
              DEFAULT_KEY_MAXSIZE));
        int numNativeBuffers = Integer.parseInt(
            getProperties().getProperty(NATIVEBUFFER_NUM_PROPERTY,
              DEFAULT_NATIVEBUFFER_NUM));
        int fieldlength = 
            Integer.parseInt(getProperties().getProperty(FIELD_LENGTH_PROPERTY, FIELD_LENGTH_PROPERTY_DEFAULT));
        int fieldcount =
            Integer.parseInt(getProperties().getProperty(FIELD_COUNT_PROPERTY, FIELD_COUNT_PROPERTY_DEFAULT));

        defClient = null;
        optMode = 1;
        client = new RdmaMemcachedClient(
                    hosts, maxKeyLength, fieldcount, (fieldlength + JSON_FORMAT_EXTRA), numNativeBuffers); 
      } else {
        // ***** using spymemcached for IP-over-IB or Eth client
        defClient = createMemcachedClient();
        checkOperationStatus = Boolean.parseBoolean(
            getProperties().getProperty(CHECK_OPERATION_STATUS_PROPERTY,
              CHECK_OPERATION_STATUS_DEFAULT));
        objectExpirationTime = Integer.parseInt(
            getProperties().getProperty(OBJECT_EXPIRATION_TIME_PROPERTY,
              DEFAULT_OBJECT_EXPIRATION_TIME));
        shutdownTimeoutMillis = Integer.parseInt(
            getProperties().getProperty(SHUTDOWN_TIMEOUT_MILLIS_PROPERTY,
              DEFAULT_SHUTDOWN_TIMEOUT_MILLIS));

        optMode = 0; 
        client = null;
      }
    } catch (Exception e) {
      throw new DBException(e);
    }
  }

  protected net.spy.memcached.MemcachedClient createMemcachedClient()
    throws Exception {
    ConnectionFactoryBuilder connectionFactoryBuilder =
        new ConnectionFactoryBuilder();

    connectionFactoryBuilder.setReadBufferSize(Integer.parseInt(
          getProperties().getProperty(READ_BUFFER_SIZE_PROPERTY,
            DEFAULT_READ_BUFFER_SIZE)));

    connectionFactoryBuilder.setOpTimeout(Integer.parseInt(
          getProperties().getProperty(OP_TIMEOUT_PROPERTY, DEFAULT_OP_TIMEOUT)));

    String protocolString = getProperties().getProperty(PROTOCOL_PROPERTY);
    connectionFactoryBuilder.setProtocol(
        protocolString == null ? DEFAULT_SPYMEMCACHED_PROTOCOL
        : ConnectionFactoryBuilder.Protocol.valueOf(protocolString.toUpperCase()));

    String failureString = getProperties().getProperty(FAILURE_MODE_PROPERTY);
    connectionFactoryBuilder.setFailureMode(
        failureString == null ? FAILURE_MODE_PROPERTY_DEFAULT
        : FailureMode.valueOf(failureString));

    // Note: this only works with IPv4 addresses due to its assumption of
    // ":" being the separator of hostname/IP and port; this is not the case
    // when dealing with IPv6 addresses.
    //
    // TODO(mbrukman): fix this.
    List<InetSocketAddress> addresses = new ArrayList<InetSocketAddress>();
    String[] hosts = getProperties().getProperty(HOSTS_PROPERTY).split(",");
    for (String address : hosts) {
      int colon = address.indexOf(":");
      int port = DEFAULT_PORT;
      String host = address;
      if (colon != -1) {
        port = Integer.parseInt(address.substring(colon + 1));
        host = address.substring(0, colon);
      }
      addresses.add(new InetSocketAddress(host, port));
    }
    return new net.spy.memcached.MemcachedClient(
        connectionFactoryBuilder.build(), addresses);
  }

  @Override
  public Status read(
      String table, String key, Set<String> fields,
      Map<String, ByteIterator> result) {
    key = createQualifiedKey(table, key);

    try {
      if (optMode == 0) {
        GetFuture<Object> future = memcachedClient().asyncGet(key);
        Object document = future.get();
        if (document != null) {
          fromJson((String) document, fields, result);
        }
        return Status.OK;
      } else {
        int futureId = client.asyncGet(key); 
        String document = client.get(futureId); 
        if (document != null) {
          fromJson((String) document, fields, result);
        }
        return Status.OK;
      }
    } catch (Exception e) {
      logger.error("Error encountered for key: " + key, e);
      e.printStackTrace(); 
      return Status.ERROR;
    }
  }

  @Override
  public Status scan(
      String table, String startkey, int recordcount, Set<String> fields,
      Vector<HashMap<String, ByteIterator>> result){
    return Status.NOT_IMPLEMENTED;
  }

  @Override
  public Status update(
      String table, String key, Map<String, ByteIterator> values) {
    key = createQualifiedKey(table, key);

    try {
      if (optMode == 0) { //default spymemcached mode
        OperationFuture<Boolean> future =
            memcachedClient().replace(key, objectExpirationTime, toJson(values));
        return getReturnCode(future);
      } else {
        int futureId = client.asyncSet(key, values, 0, 0); 
        return client.status(futureId); 
      }
    } catch (Exception e) {
      logger.error("Error updating value with key: " + key, e);
      return Status.ERROR;
    }
  }

  @Override
  public Status insert(
      String table, String key, Map<String, ByteIterator> values) {
    key = createQualifiedKey(table, key);

    try {
      if (optMode == 0) { //default spymemcached mode
        OperationFuture<Boolean> future =
            memcachedClient().add(key, objectExpirationTime, toJson(values));
        return getReturnCode(future);
      } else {
        int futureId = client.asyncSet(key, values, 0, 0); 
        return client.status(futureId); 
      }
    } catch (Exception e) {
      logger.error("Error inserting value", e);
      e.printStackTrace(); 
      return Status.ERROR;
    }
  }

  @Override
  public Status delete(String table, String key) {
    key = createQualifiedKey(table, key);

    try {
      if (optMode == 0) { //default spymemcached mode
        OperationFuture<Boolean> future = defClient.delete(key);
        return getReturnCode(future);
      } else {
        return client.delete(key, key.length()); 
      }
    } catch (Exception e) {
      logger.error("Error deleting value", e);
      return Status.ERROR;
    }
  }

  protected Status getReturnCode(OperationFuture<Boolean> future) {
    if (!checkOperationStatus) {
      return Status.OK;
    }
    if (future.getStatus().isSuccess()) {
      return Status.OK;
    } else if (TEMPORARY_FAILURE_MSG.equals(future.getStatus().getMessage())) {
      return new Status("TEMPORARY_FAILURE", TEMPORARY_FAILURE_MSG);
    } else if (CANCELLED_MSG.equals(future.getStatus().getMessage())) {
      return new Status("CANCELLED_MSG", CANCELLED_MSG);
    }
    return new Status("ERROR", future.getStatus().getMessage());
  }

  @Override
  public void cleanup() throws DBException {
    if (optMode == 0) {
      memcachedClient().shutdown(shutdownTimeoutMillis, MILLISECONDS);
    } else {
      client.cleanup(); 
    }
  }

  protected static String createQualifiedKey(String table, String key) {
    return MessageFormat.format("{0}-{1}", table, key);
  }

  protected static void fromJson(
      String value, Set<String> fields,
      Map<String, ByteIterator> result) throws IOException {
    JsonNode json = MAPPER.readTree(value);
    boolean checkFields = fields != null && !fields.isEmpty();
    for (Iterator<Map.Entry<String, JsonNode>> jsonFields = json.getFields();
        jsonFields.hasNext();
        /* increment in loop body */) {
      Map.Entry<String, JsonNode> jsonField = jsonFields.next();
      String name = jsonField.getKey();
      if (checkFields && fields.contains(name)) {
        continue;
      }
      JsonNode jsonValue = jsonField.getValue();
      if (jsonValue != null && !jsonValue.isNull()) {
        result.put(name, new StringByteIterator(jsonValue.asText()));
      }
    }
  }

  protected static String toJson(Map<String, ByteIterator> values)
    throws IOException {
    ObjectNode node = MAPPER.createObjectNode();
    Map<String, String> stringMap = StringByteIterator.getStringMap(values);
    for (Map.Entry<String, String> pair : stringMap.entrySet()) {
      node.put(pair.getKey(), pair.getValue());
    }
    JsonFactory jsonFactory = new JsonFactory();
    Writer writer = new StringWriter();
    JsonGenerator jsonGenerator = jsonFactory.createJsonGenerator(writer);
    MAPPER.writeTree(jsonGenerator, node);
    return writer.toString();
  }
}
