/* Copyright (c) 2011-2017, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the OSU HiBD-Benchmarks software package
 * developed by the team members of The Ohio State University's
 * Network-Based Computing Laboratory (NBCL), headed by Professor
 * Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to
 * the copyright file COPYRIGHT in the top level directory.
 *
 */
/**
 * Copyright (c) 2014-2015 YCSB contributors. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * permissions and limitations under the License. See accompanying
 * LICENSE file.
 */

package edu.osu.hibd.ohb.memcached; 

import com.yahoo.ycsb.Status;
import com.yahoo.ycsb.ByteIterator;
import com.yahoo.ycsb.StringByteIterator;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.Map;
import java.io.StringWriter;
import java.io.Writer;

import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonGenerator;
import org.codehaus.jackson.node.ObjectNode;
import org.codehaus.jackson.map.ObjectMapper;

/**
 * RDMA Libmemcached client implementation.
 */
public class RdmaMemcachedClient {

  public static final String RDMA = "RDMA"; 
  private static final int MEMCACHED_SUCCESS = 0; 
  private static final int MEMCACHED_FAILURE = 1;
  private static int numClients = 0; 
  protected static final ObjectMapper MAPPER = new ObjectMapper();

  private long objectId; 
  private String hostList; 
  private int maxKeyLength; 
  private int recordLength; 
  private int nBuffer; 
  private int bufLength; 
  private ByteBuffer bigBuffer;

  static {
    try {
      String ycsbHome = System.getenv("YCSB_HOME"); 
      String ycsbNativeLib = new String(ycsbHome + "/memcached/target/native_libs/librdmamemcycsb.so");
      System.load(ycsbNativeLib);
    } catch(UnsatisfiedLinkError e) {
      System.out.println("No Load native library for YCSB Rdma Memcached.");
      e.printStackTrace();
    }
  }

  public long objectId() {
    return objectId; 
  }

  public RdmaMemcachedClient(
          String hosts, int maxkeylen, int fieldcount, int fieldlen, int nbuffers) { 
    hostList = hosts;
    objectId = this.hashCode();
    maxKeyLength = maxkeylen; 
    recordLength = (fieldlen * fieldcount); 
    nBuffer = nbuffers; 
    // *** [ KEY | 0 | VALUE | 0 ] [....]
    bufLength = (fieldlen * fieldcount) + 2 /* null */ +  maxKeyLength;  
    bigBuffer = ByteBuffer.allocateDirect(bufLength * nBuffer); 

    int rc = memcachedCreate(objectId, hostList, 
            bufLength, nBuffer, (maxKeyLength+1) /*key offset*/, bigBuffer); 
    if (rc != MEMCACHED_SUCCESS) {
      this.objectId = -1;  
    } else {
      numClients++; 
    }
  }

  public Status delete(String key, int keylen) {
    int rc = memcachedDelete(this.objectId, key, keylen); 

    if (rc == MEMCACHED_SUCCESS) {
      return Status.OK;
    } else {
      return Status.ERROR;
    }
  }

  public int asyncSet(String key, Map<String, ByteIterator> values,
      int expiration, int flags) {
    int rc = MEMCACHED_SUCCESS; 

    int requestId = getFreeRequest(this.objectId);    
    if (requestId == -1) {
      System.out.println("here"); 
      rc = MEMCACHED_FAILURE; 
    } else {
      int offset = requestId * this.bufLength; 

      // *** populate key and value into direct buffer
      int keylen = keyToByteBuffer(bigBuffer, offset, key); 
      int vallen = valueToByteBuffer(
          this.bigBuffer, offset + (this.maxKeyLength + 1), values); 

      rc = memcachedIset(this.objectId, keylen, vallen, requestId, 
               expiration, flags);
    }

    if (rc == MEMCACHED_SUCCESS) {
      return requestId;
    } else {
      return -1;
    }
  }

  public int asyncGet(String key) {
    int rc = MEMCACHED_SUCCESS; 

    int requestId = getFreeRequest(this.objectId);    
    if (requestId == -1) {
      rc = MEMCACHED_FAILURE; 
    } else {
      int offset = requestId * this.bufLength;
      // *** populate key 
      int keylen = keyToByteBuffer(bigBuffer, offset, key); 
    
      rc = memcachedIget(this.objectId, keylen, requestId); 
    }

    if (rc == MEMCACHED_SUCCESS) {
      return requestId;
    } else {
      return -1;
    }
  }

  public Status status(int requestId) {
    int rc = memcachedWait(this.objectId, requestId); 

    if (rc == MEMCACHED_SUCCESS) { 
      return Status.OK;
    } else {
      return Status.ERROR; 
    }
  }

  public String get(int requestId) {
    int vallen = memcachedGetWait(this.objectId, requestId); 
    int valoffset = 
        (requestId * this.bufLength) + this.maxKeyLength + 1; 

    if (vallen > 0) { //*** successful get
      return byteBufferToStr(bigBuffer, valoffset, vallen);
    } else {
      return null; 
    }
  }

  public void cleanup() {
    memcachedFree(this.objectId); 
  }

  protected static String byteBufferToStr(
          ByteBuffer buf, int offset, int len) {
    buf.position(offset); 

    try {
      if (buf.hasArray()) { 
        return new String(buf.array(), 0, len, Charset.forName("UTF-8")); 
      } else {
        byte [] dst = new byte[len];
        buf.get(dst, 0, len); 

        return new String(dst, 0, len, Charset.forName("UTF-8"));
      }
    } catch(Exception e) {
      e.printStackTrace(); 
      return null;
    }
  }

  protected static int valueToByteBuffer(
          ByteBuffer buf, int offset, Map<String, ByteIterator> values) {
    buf.position(offset); 

    try {
      ObjectNode node = MAPPER.createObjectNode();
      Map<String, String> stringMap = StringByteIterator.getStringMap(values);
      for (Map.Entry<String, String> pair : stringMap.entrySet()) {
        node.put(pair.getKey(), pair.getValue());
      }
      JsonFactory jsonFactory = new JsonFactory();
      Writer writer = new StringWriter();
      JsonGenerator jsonGenerator = jsonFactory.createJsonGenerator(writer);
      MAPPER.writeTree(jsonGenerator, node);
      buf.put(writer.toString().getBytes("UTF-8")); 
    } catch(IOException e) {
      e.printStackTrace();
    }

    return buf.position() - offset; 
  }

  protected static int keyToByteBuffer(
          ByteBuffer buf, int offset, String key) {
    buf.position(offset); 
    buf.put(key.getBytes());

    return key.length(); 
  }

  //***** NATIVE APIs *******
  public native int memcachedCreate(long objId, String servers, 
          long buflen, int nbufs, int off, ByteBuffer b); 
  public native int memcachedFree(long objId); 
  public native int getFreeRequest(long objId);

  public native int memcachedDelete(long objId, String key, int klen); 
  public native int memcachedIset(
      long objId, int klen, long vlen, int reqId, int exp, int flags); 
  public native int memcachedIget(long objId, int klen, int reqId);
  // *** return request length - block for completion 
  public native int memcachedGetWait(long objId, int reqId); 
  // *** return request status - block for completion
  public native int memcachedWait(long objId, int reqId);
  
  //***** NATIVE APIs *******
}
