/* Copyright (c) 2011-2017, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the OSU HiBD-Benchmarks software package
 * developed by the team members of The Ohio State University's
 * Network-Based Computing Laboratory (NBCL), headed by Professor
 * Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to
 * the copyright file COPYRIGHT in the top level directory.
 *
 */

#include <jni.h>
#include "edu_osu_hibd_ohb_memcached_RdmaMemcachedClient.h"
#include <libmemcached/memcached.h>
#include <stdio.h>
#include <time.h>       
#include <map>

#define YCSB_ACCESS_DONE 255

using namespace std; 

/******* dbg *********/

#if 0
  #define dbg(fmt, ...) do{ \
          fprintf(stderr, "[ %s ] : " fmt, __func__, __VA_ARGS__); fflush(stderr); } while(0)
#else
  #define  dbg(fmt, ...)
#endif
/******* dbg *********/

typedef struct {
    char *buffer;
    int num_buffers;
    long buffer_length;  
    int value_offset; 
    memcached_request_st **request_list; 
    int current_indx; 
} native_buffer_t; 

std::map<long, native_buffer_t*> buffer_map; 
std::map<long, memcached_st*> client_map; 
std::map<long, memcached_request_st*> request_map; 

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedCreate
(JNIEnv *env, jobject obj, jlong oid, jstring server_list, jlong buflen, jint nbufs, jint voff, jobject buf) 
{
    memcached_st *memc= memcached_create(NULL);
    const char *opt_server = env->GetStringUTFChars(server_list, 0);
    if (!opt_server || !memc) {
        fprintf(stderr, "Empty server list provided..(%s)\n", opt_server);
        return 1;
    } 
    memcached_server_st* servers= memcached_servers_parse(opt_server);
    memcached_return rc = memcached_server_push(memc, servers);
    if (servers == NULL || memcached_server_list_count(servers) == 0 || rc != MEMCACHED_SUCCESS)
    {
        fprintf(stderr, "Invalid server list provided..(%s)\n", opt_server);
        return 1;
    }

    dbg("Connected to server: %s (oid %lu)\n", opt_server, oid); 
    client_map[oid] = memc; 

    native_buffer_t *b = (native_buffer_t *) malloc(sizeof(native_buffer_t)); 
    if (!b) return 1; 

    char *address = (char*) env->GetDirectBufferAddress(buf); 
    long capacity = env->GetDirectBufferCapacity(buf); 

    if (!address || (address && capacity < (buflen * nbufs))) { 
        return 1;
    } else {
        b->buffer_length = buflen; 
        b->num_buffers = nbufs; 
        b->buffer = address; 
        b->value_offset = voff; 
        b->current_indx = 0;
        buffer_map[oid] = b; 

        b->request_list = (memcached_request_st **)malloc(sizeof(memcached_request_st *) * nbufs); 
        if (!b->request_list) {
            free(address);
            return 1;
        }
        for (int i = 0; i < nbufs; i++) {
            //*** create place-holder requests and init to YCSB_ACCESS_DONE for our reference
            //*** ensure to update req->value and req->value_length before issuing bset/bget
            //*** to avoid errors 
            b->request_list[i] = memcached_request_create(memc, address + i * buflen, 0, 0); 
            b->request_list[i]->response = YCSB_ACCESS_DONE; 
        }

        dbg("Created buffer capacity %ld (oid %lu)\n", buflen * nbufs, oid); 
        return 0;
    }
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedFree
(JNIEnv *env, jobject obj, jlong oid) 
{
    memcached_st *memc = client_map[oid];
    if (!memc) {
        return 1; 
    } else { 
        native_buffer_t *b = buffer_map[oid]; // *** free native buffer 
        if (!b) return 1;
        if (b->request_list) free(b->request_list); 
        if (b->buffer) free(b->buffer); 

        free(b); 
        memcached_free(memc); 

        return 0;
    }
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_getFreeRequest
(JNIEnv *env, jobject obj, jlong oid) {
    native_buffer_t *b = buffer_map[oid];
    if (!b) return -1;

    int reqid = b->current_indx; 
    b->current_indx = (b->current_indx + 1) % b->num_buffers; 

    memcached_request_st *req = b->request_list[reqid];
    //*** Ensure that request is free to use. Since we currently support only 
    //*** blocking semantics at the YCSB-Java later, we expect buffer access from Java layer 
    //*** to be completed. TO-DO: support for async operations
    if (YCSB_ACCESS_DONE != req->response) { 
        return -1; //*** index unavailable                           
    }

    return reqid; 
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedDelete
(JNIEnv *env, jobject obj, jlong oid, jstring keystr, jint keylen) 
{
    memcached_st *memc = client_map[oid];
    const char *key = env->GetStringUTFChars(keystr, 0);
    if (!memc || !key) return 1; 
   
    memcached_return_t rc = memcached_delete(memc, key, keylen, 0); 
    env->ReleaseStringUTFChars(keystr, key);

    if (rc == MEMCACHED_SUCCESS) { 
        return 0;
    } else {
        return 1;
    }
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedIset
(JNIEnv *env, jobject obj, jlong oid, jint keylen, jlong vallen, jint reqid, jint expire, jint flags) {
    memcached_return_t rc = MEMCACHED_FAILURE;
    memcached_st *memc = client_map[oid];
    native_buffer_t *b = buffer_map[oid]; 
    if (!b || !keylen || !vallen || !memc) return 1;

    memcached_request_st *req = b->request_list[reqid]; 
    if (!req || (keylen+vallen+2) > b->buffer_length) return 1;

    char *key = (char *)b->buffer + reqid * b->buffer_length; 
    req->value = (char *)key + b->value_offset;
    req->value_length = vallen; 

    rc= memcached_bset(memc, key, keylen, req, (time_t)expire, (uint32_t)flags);
    if(rc == MEMCACHED_SUCCESS) {
        dbg("Key store request %d successful: %.*s %lu %d\n", reqid, keylen, key, vallen, keylen);
        dbg("Value %lu: %.*s\n", vallen, (int)vallen, (char*)req->value);
        return 0;
    } else {
        dbg("Key store request failed: %.*s %lu (%s)\n", 
                keylen, key, vallen, memcached_strerror(memc, rc));
        return 1; 
    }
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedIget
(JNIEnv *env, jobject obj, jlong oid, jint keylen, jint reqid) {
    memcached_return_t rc = MEMCACHED_FAILURE;
    memcached_st *memc = client_map[oid];
    native_buffer_t *b = buffer_map[oid]; 
    if (!b || !keylen || !memc) return 1;

    memcached_request_st *req = b->request_list[reqid]; 
    if (!req) return 1;

    char *key = (char *)b->buffer + reqid * b->buffer_length; 
    req->value = (char *)key + b->value_offset;
    req->value_length = 0; 

    rc= memcached_bget(memc, key, keylen, req);
    if (rc == MEMCACHED_SUCCESS) {
        dbg("Key fetch request successful: %.*s %lu\n", keylen, key, (long)((char*)req->value - b->buffer));
        return 0;
    } else {
        dbg("Key store failed: %.*s (%s)\n", keylen, key, memcached_strerror(memc, rc));
        return 1;
    }
}


JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedWait
(JNIEnv *env, jobject obj, jlong oid, jint reqid) {
    memcached_st *memc = client_map[oid];
    native_buffer_t *b = buffer_map[oid]; 
    if (!b || !memc) return 1;

    memcached_request_st *req = b->request_list[reqid]; 
    if (!req) return 1;

    memcached_return_t rc = memcached_wait(memc, req);
    if (rc == MEMCACHED_SUCCESS) {
        dbg("Request %d completed successfully: vallen %lu\n", reqid, req->value_length);
        req->response = YCSB_ACCESS_DONE;
        return 0;
    } else {
        dbg("Request %d completed with failure: %s\n", reqid, memcached_strerror(memc, rc));
        return 1;
    }
}

JNIEXPORT jint JNICALL Java_edu_osu_hibd_ohb_memcached_RdmaMemcachedClient_memcachedGetWait
(JNIEnv *env, jobject obj, jlong oid, jint reqid) {
    memcached_st *memc = client_map[oid];
    native_buffer_t *b = buffer_map[oid]; 
    if (!b || !memc) return 0;

    memcached_request_st *req = b->request_list[reqid]; 
    if (!req) return 0;

    memcached_return_t rc = memcached_wait(memc, req);
    if (rc == MEMCACHED_SUCCESS) {
        dbg("Request %d completed successfully: vallen %lu\n", reqid, req->value_length);
        dbg("Value %.*s\n", (int)req->value_length, (char*)req->value);
        req->response = YCSB_ACCESS_DONE;
        return req->value_length; 
    } else {
        dbg("Request %d completed with failure: %s\n", reqid, memcached_strerror(memc, rc));
        return 0; 
    }
}
